--[[ LootSorter
     /ls

     What are the four lowercase letters that are not legal flag arguments
     to the Berkeley UNIX version of `ls'?

     (Answer:  Actually, there's five, ejvyz.)

     ls.windows = {
       { window =,
	 filter =,
	 selected =,
	 order =,
	 order_asc =,
       }
       [...]
     }
     ls.spare_windows = {
       window,
       ...
     }

]]--

local ls = {}
ls.version = "0.12-130712-20:34:39"
local lbag = Library.LibBaggotry
local filt = Library.LibEnfiltrate
LootSorter = ls

ls.item_lines = 20
ls.windows = {}
ls.spare_windows = {}
ls.layer = 0

function ls.printf(fmt, ...)
  print(string.format(fmt or 'nil', ...))
end

function ls.variables_loaded(event, name)
  if name == 'LootSorter' then
    LootSorterAccount = LootSorterAccount or { window_x = 150, window_y = 150 }
    ls.account_vars = LootSorterAccount
  end
end

function ls.display_loc(frame, item)
  local loc
  if item._slotspec then
    local s = lbag.slotspec_p(item._slotspec)
    if s then
      local type, p1, p2 = Utility.Item.Slot.Parse(item._slotspec)
      loc = type
    else
      loc = item._slotspec
    end
  else
    loc = "<Unknown>"
  end
  frame:SetText(loc)
  frame:SetFontColor(0.98, 0.98, 0.98)
end

function ls.display_icon(frame, item)
  if not item then
    return
  end
  if item.icon then
    frame:SetTexture("Rift", item.icon)
    frame:SetVisible(true)
  else
    frame:SetVisible(false)
  end
end

function ls.display_name(frame, item)
  frame:SetText(item.name or "NO NAME")
  local r, g, b = lbag.rarity_color(item.rarity)
  frame:SetFontColor(r, g, b)
end

function ls.display_owner(frame, item)
  local owner, suffix
  owner = item._character or "--"
  owner = string.upper(string.sub(owner, 1, 1)) .. string.sub(owner, 2)
  frame:SetText(owner)
  if owner == '--' then
    frame:SetFontColor(0.5, 0.5, 0.5)
  else
    frame:SetFontColor(0.98, 0.98, 0.98)
  end
end

function ls.display_rarity(frame, item)
  frame:SetText(item.rarity or "common")
  local r, g, b = lbag.rarity_color(item.rarity)
  frame:SetFontColor(r, g, b)
end

function ls.display_level(frame, item)
  local x
  local me = Inspect.Unit.Detail('player')
  if item.requiredLevel then
    x = tostring(item.requiredLevel)
    if me and me.level and (me.level < item.requiredLevel) then
      frame:SetFontColor(0.9, 0, 0)
    else
      frame:SetFontColor(0.98, 0.98, 0.98)
    end
  else
    x = "--"
    frame:SetFontColor(0.5, 0.5, 0.5)
  end
  frame:SetText(x)
end

function ls.display_qty(frame, item)
  local x
  if item.stackMax or (item.stack and item.stack > 1) then
    x = string.format("%d/%d", item.stack, item.stackMax or 1)
  else
    x = ""
  end
  frame:SetText(x)
  frame:SetFontColor(0.98, 0.98, 0.98)
end

function ls.order_generic(window, acmp, bcmp, invert)
  if acmp == bcmp then
    return false
  end
  local c = acmp < bcmp
  if invert then
    c = not c
  end
  if c then
    return window.order_asc
  else
    return not window.order_asc
  end
end

function ls.order_qty_calc(item)
  if not item then
    return 0
  end
  if item.stackMax then
    return item.stack or 1
  else
    return 0
  end
end

function ls.order_level_calc(item)
  if not item then
    return -1
  end
  return item.requiredLevel or -1
end

function ls.order_level(window, a, b)
  local acmp, bcmp
  a = window.item_list[a]
  b = window.item_list[b]
  acmp = ls.order_level_calc(a)
  bcmp = ls.order_level_calc(b)
  return ls.order_generic(window, acmp, bcmp, true)
end


function ls.order_qty(window, a, b)
  local acmp, bcmp
  local ai = window.item_list[a]
  local bi = window.item_list[b]
  acmp = ls.order_qty_calc(ai)
  bcmp = ls.order_qty_calc(bi)
  return ls.order_generic(window, acmp, bcmp, true)
end

function ls.order_owner(window, a, b)
  local acmp, bcmp
  a = window.item_list[a]
  b = window.item_list[b]
  acmp = a and (a._character or "") or ""
  bcmp = b and (b._character or "") or ""
  return ls.order_generic(window, acmp, bcmp)
end

function ls.order_name(window, a, b)
  local acmp, bcmp
  a = window.item_list[a]
  b = window.item_list[b]
  acmp = a and (a.name or "") or ""
  bcmp = b and (b.name or "") or ""
  return ls.order_generic(window, acmp, bcmp)
end

function ls.order_rarity(window, a, b)
  local acmp, bcmp
  a = window.item_list[a]
  b = window.item_list[b]
  acmp = a and (lbag.rarity_p(a.rarity)) or 0
  bcmp = b and (lbag.rarity_p(b.rarity)) or 0
  return ls.order_generic(window, acmp, bcmp, true)
end

function ls.order_loc(window, a, b)
  local acmp, bcmp
  a = window.item_list[a]
  b = window.item_list[b]
  acmp = a and (a._slotspec) or ""
  bcmp = b and (b._slotspec) or ""
  return ls.order_generic(window, acmp, bcmp)
end

ls.order = 'name'

ls.columns = {
  icon = { name = 'Icn', width = 20, type = "Texture", display = ls.display_icon, order = nil, },
  name = { name = 'Name', width = 200, display = ls.display_name, order = ls.order_name },
  qty = { name = 'Qty', width = 70, display = ls.display_qty, order = ls.order_qty },
  level = { name = 'Level', width = 70, display = ls.display_level, order = ls.order_level },
  rarity = { name = 'Rarity', width = 75, display = ls.display_rarity, order = ls.order_rarity },
  loc = { name = 'Location', width = 100, display = ls.display_loc, order = ls.order_loc },
  owner = { name = 'Owner', width = 150, display = ls.display_owner, order = ls.order_owner },
}

ls.column_order = {
  'icon', 'name', 'qty', 'level', 'rarity', 'loc', 'owner'
}

function ls.makecolumns(window, item)
  local xoff = 0
  for i, v in ipairs(ls.column_order) do
    local f = item.subframes[i]
    local data = ls.columns[v]
    f:SetPoint("TOPLEFT", item.frame, "TOPLEFT", xoff, 0)
    f:SetPoint("BOTTOMRIGHT", item.frame, "BOTTOMLEFT", xoff + data.width, 0)
    f:SetMouseMasking("limited")
    f:EventAttach(Event.UI.Input.Mouse.Left.Click, function(event, ...) ls.leftclick(window, item.index, v, ...) end, "left click")
    f:EventAttach(Event.UI.Input.Mouse.Right.Click, function(event, ...) ls.rightclick(window, item.index, v, ...) end, "right click")
    f:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(event, ...) ls.mousein(window, item.index, v, ...) end, "mouse in")
    f:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(event, ...) ls.mouseout(window, item.index, v, ...) end, "mouse out")
    xoff = xoff + data.width + 5
  end
end

function ls.mousein(window, idx, subframe, ...)
  -- ls.printf("mousein %d %s", idx or -1, subframe and tostring(subframe) or 'nil')
  if idx then
    local pos = math.floor(window.scrollbar:GetPosition())
    local item_idx = idx + pos
    local item_name = window.item_ordered[item_idx]
    if window.item_list[item_name] and window.item_list[item_name].type then
      -- sometimes Command.Tooltip throws spurious errors
      local foo = function() Command.Tooltip(window.item_list[item_name].type) end
      pcall(foo)
    end
  end
end

function ls.mouseout(window, idx, ...)
  Command.Tooltip(nil)
end

function ls.makeitem(window, idx, no_events, complete)
  local item = {}
  local columns
  item.index = idx
  item.frame = UI.CreateFrame("Frame", "Item Frame", window.window)
  item.subframes = {}
  if not no_events then
    item.frame:EventAttach(Event.UI.Input.Mouse.Left.Click, function(event, ...) ls.leftclick(window, idx, nil, ...) end, "left click")
    item.frame:EventAttach(Event.UI.Input.Mouse.Cursor.In, function(event, ...) ls.mousein(window, idx, 'outer', ...) end, "mouse in")
    item.frame:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function(event, ...) ls.mouseout(window, idx, 'outer', ...) end, "mouse out")
  end

  if complete then
    columns = ls.columns
  else
    columns = ls.column_order
  end
  for i, v in pairs(columns) do
    if type(v) == 'string' then
      v = ls.columns[v]
    end
    if idx == 0 then
      item.subframes[i] = UI.CreateFrame("Text", "Column", item.frame)
    else
      item.subframes[i] = UI.CreateFrame(v.type or "Text", "Column", item.frame)
    end
  end
  return item
end

function ls.setgrey(frame, count)
  local grey = 0.25
  frame:SetBackgroundColor(grey, grey, grey, 0.4)
end

function ls.leftclick(window, idx, field, ...)
  if idx == 0 then
    local func = ls.columns[field] and ls.columns[field].order
    -- if there's no order function, we don't care
    if func then
      if window.order == func then
        window.order_asc = not window.order_asc
      else
        window.order_asc = true
	window.order = func
      end
      ls.reorder(window)
      ls.show_items(window)
    end
    for i, v in pairs(ls.columns) do
      if i == field then
        window.heading.subframes[i]:SetFontColor(1, 1, 0.6, 1.0)
      else
        window.heading.subframes[i]:SetFontColor(0.8, 0.8, 0.8, 1.0)
      end
    end
  elseif idx then
    local pos = math.floor(window.scrollbar:GetPosition())
    local item = window.item_ordered[idx + pos]
    if item then
      slot, owner = lbag.slotspec_p(item)
      if slot and owner == lbag.whoami() then
        Command.Item.Standard.Left(slot)
      end
    end
    window.selected_slot = item
    ls.show_items(window)
  end
end

function ls.rightclick(window, idx, field, ...)
  if idx == 0 then
    local func = ls.columns[field] and ls.columns[field].order
    -- if there's no order function, we don't care
    if func then
      if window.order == func then
        window.order_asc = not window.order_asc
      else
        window.order_asc = true
	window.order = func
      end
      ls.reorder(window)
      ls.show_items(window)
    end
    for i, v in pairs(ls.columns) do
      if i == field then
        window.heading.subframes[i]:SetFontColor(1, 1, 0.6, 1.0)
      else
        window.heading.subframes[i]:SetFontColor(0.8, 0.8, 0.8, 1.0)
      end
    end
  elseif idx then
    local pos = math.floor(window.scrollbar:GetPosition())
    local item = window.item_ordered[idx + pos]
    if item then
      slot, owner = lbag.slotspec_p(item)
      if slot and owner == lbag.whoami() then
        Command.Item.Standard.Right(slot)
      end
    end
    window.selected_slot = item
  end
end

-- handle dragging

function ls.mousemove(window, x, y)
  ls.account_vars.window_x = x
  ls.account_vars.window_y = y
end

local window_template = {}

function window_template:close()
  remove_me = nil
  for idx, window in ipairs(ls.windows) do
    if window == self then
      remove_me = idx
    end
  end
  if remove_me then
    table.remove(ls.windows, remove_me)
  end
  -- don't keep an old selection around
  self.selected_slot = nil
  table.insert(ls.spare_windows, self)
  self.window:SetVisible(false)
end

function window_template:get()
  if ls.spare_windows[1] then
    local win = ls.spare_windows[1]
    table.remove(ls.spare_windows, 1)
    return win
  else
    return window_template:new()
  end
end

function window_template:new()
  local o = {
    order = 'name',
    order_asc = true
  }
  setmetatable(o, self)
  self.__index = self

  o.window = UI.CreateFrame("RiftWindow", "LootSorter", ls.ui)
  o.window:SetWidth(800)
  o.window:SetTitle("LootSorter")
  o.window:SetPoint("TOPLEFT", UIParent, "TOPLEFT", ls.account_vars and ls.account_vars.window_x or 150, ls.account_vars and ls.account_vars.window_y or 150)

  local l, t, r, b = o.window:GetTrimDimensions()
  r = r * -1
  b = b * -1

  o.closebutton = UI.CreateFrame("RiftButton", "LootSorter", o.window)
  o.closebutton:SetSkin("close")
  o.closebutton:SetPoint("TOPRIGHT", o.window, "TOPRIGHT", r + 5, 15)
  o.closebutton:EventAttach(Event.UI.Input.Mouse.Left.Click, function() o:close() end, "close")

  o.scrollbar = UI.CreateFrame("RiftScrollbar", "LootSorter", o.window)
  o.scrollbar:SetPoint("TOPRIGHT", o.window, "TOPRIGHT",r - 2, t + 80)
  o.scrollbar:SetPoint("BOTTOMRIGHT", o.window, "BOTTOMRIGHT", r - 2, b - 2)
  -- only active when there is scrolletry to do
  o.scrollbar:SetEnabled(false)
  o.scrollbar:SetRange(0, 1)
  o.scrollbar:SetPosition(0)
  o.scrollbar:EventAttach(Event.UI.Scrollbar.Change, function() ls.show_items(o) end, "scrollbar change")
  o.window:EventAttach(Event.UI.Input.Mouse.Wheel.Back, function() o.scrollbar:Nudge(3) end, "wheel back")
  o.window:EventAttach(Event.UI.Input.Mouse.Wheel.Forward, function() o.scrollbar:Nudge(-3) end, "wheel forward")
  local w = o.scrollbar:GetWidth()

  o.window:GetContent():SetMouseMasking("full")
  Library.LibDraggable.draggify(o.window, o.mousemove)

  -- "selected" is where we show the selected item.
  o.selected = ls.makeitem(o, nil, true, true)
  for key, frame in pairs(o.selected.subframes) do
    if key == 'icon' then
      frame:SetPoint("TOPLEFT", o.window, "TOPLEFT", l + 4, t + 5)
      frame:SetPoint("BOTTOMRIGHT", o.window, "TOPLEFT", l + 4 + 35, t + 40)
    elseif key == 'name' then
      frame:SetPoint("TOPLEFT", o.window, "TOPLEFT", l + 40, t + 5)
    elseif key == 'qty' then
      frame:SetPoint("TOPLEFT", o.window, "TOPLEFT", l + 40, t + 25)
    elseif key == 'level' then
      frame:SetPoint("TOPLEFT", o.window, "TOPLEFT", l + 240, t + 5)
    elseif key == 'rarity' then
      frame:SetPoint("TOPLEFT", o.window, "TOPLEFT", l + 240, t + 25)
    elseif key == 'loc' then
      frame:SetPoint("TOPLEFT", o.window, "TOPLEFT", l + 340, t + 25)
    elseif key == 'owner' then
      frame:SetPoint("TOPLEFT", o.window, "TOPLEFT", l + 340, t + 5)
    end
  end

  o.items = {}
  for i = 0, ls.item_lines do
    o.items[i] = ls.makeitem(o, i)
    o.items[i].index = i
    o.items[i].frame:SetPoint("BOTTOMRIGHT", o.window, "BOTTOMRIGHT",
	r - w - 2,
	b - 20 * (ls.item_lines - i) - 2)
    o.items[i].frame:SetWidth(800 - w - l + r - 5)
    o.items[i].frame:SetHeight(18)
    ls.makecolumns(o, o.items[i])
    if i == 0 then
      o.heading = o.items[i]
      o.heading.frame:SetBackgroundColor(0, 0, 0, 0)
      ls.show_item(o.heading, {}, true)
    else
      ls.setgrey(o.items[i].frame, i)
    end
  end

  return o
end

function ls.show_item(frame, item, heading)
  local columns
  if frame.subframes[1] then
    columns = ls.column_order
  else
    columns = ls.columns
  end
  for key, value in pairs(columns) do
    if type(value) == 'string' then
      value = ls.columns[value]
    end
    local sub = frame.subframes[key]
    if not sub then
      ls.printf("Error:  Missing subframe %s", tostring(key))
      break
    end
    if heading then
      sub:SetText(value.name)
      sub:SetFontColor(0.8, 0.8, 0.8)
    else
      local func = value and value.display
      if func then
	if item then
          func(sub, item)
	  sub:SetVisible(true)
	else
	  sub:SetVisible(false)
	end
      else
        sub:SetText(data.name)
	sub:SetFontColor(1, 0.3, 0.3)
      end
    end
  end
  frame.frame:SetVisible(true)
end

function ls.show_selected(window, item)
  ls.show_item(window.selected, item, false)
end

function ls.show_items(window)
  local max
  local pos = math.floor(window.scrollbar:GetPosition())

  local selected_slot, selected_item

  if window.selected_slot then
    selected_item = window.item_list[window.selected_slot]
  end

  ls.show_selected(window, selected_item)

  for i = 1, ls.item_lines do
    local item_idx = i + pos
    local item_name = window.item_ordered[item_idx]
    if item_name then
      local item = window.item_list[item_name]
      if not item then
	ls.printf("Trying to look at pos %d out of %d, got %s",
	  item_idx, window.item_count, item_name or "nil")
      else
        ls.show_item(window.items[i], item)
      end
      if item_name == window.selected_slot then
        window.items[i].frame:SetBackgroundColor(0.4, 0.4, 0.2, 0.4)
      else
        ls.setgrey(window.items[i].frame, item_idx)
      end
    else
      window.items[i].frame:SetVisible(false)
    end
  end
  window.window:SetVisible(true)
end

function ls.reorder(window)
  local func = nil
  if window.order and ls.columns[window.order] and ls.columns[window.order].order then
    func = function(...) return ls.columns[window.order].order(window, ...) end
  end
  table.sort(window.item_ordered, func)
end

function ls.refresh()
  for _, window in ipairs(ls.windows) do
    if window.window:GetVisible() then
      ls.dump(window)
    end
  end
end

function ls.dump(window, newfilter)
  if not window then
    for _, win in pairs(ls.windows) do
      if win.filter == newfilter then
        window = win
      end
    end
    if not window then
      window = window_template:get()
      table.insert(ls.windows, window)
      ls.layer = ls.layer + 1
      window.window:SetLayer(ls.layer)
    end
    if window and newfilter then
      window.filter = newfilter
    end
  elseif newfilter then
    window.filter = newfilter
  end
  if not window.filter then
    ls.printf("ls.dump(%s, %s): no window.filter",
      tostring(window), tostring(filter))
    return
  end
  window.item_list = lbag.expand(window.filter)

  window.item_count = 0
  window.item_ordered = {}
  for k, v in pairs(window.item_list) do
    window.item_count = window.item_count + 1
    table.insert(window.item_ordered, k)
  end
  ls.reorder(window)

  local max = window.item_count - ls.item_lines
  local _, sbmax = window.scrollbar:GetRange()
  local relative = window.scrollbar:GetPosition() / sbmax
  -- ls.printf("relative position: %d/%d => %f", window.scrollbar:GetPosition(), sbmax, relative)
  if relative > 1 then
    relative = 1
  end
  if max > 0 then
    window.scrollbar:SetEnabled(true)
    window.scrollbar:SetRange(0, max)
    window.scrollbar:SetThickness(ls.item_lines)
  else
    max = 0
    window.scrollbar:SetEnabled(false)
    window.scrollbar:SetRange(0, 1)
    window.scrollbar:SetThickness(1)
  end
  window.scrollbar:SetPosition(math.floor((relative * max) + 0.5))
  ls.show_items(window)
end

function ls.slashcommand(args)
  local dump = false
  local temporary = false
  local created = false
  if not args then
    return
  end

  if args.v then
    ls.printf("version %s", ls.version)
    return
  end

  if args.f then
    local filter = filt.Filter:load(args.f, 'LootSorter')
    if not filter then
      filter = filt.Filter:new(args.f, 'item', 'LootSorter')
      created = true
    end
    args.f = nil
  else
    filter = filt.Filter:new(nil, 'item', 'LootSorter')
    temporary = true
  end

  local changed = false
  if lbag.apply_args(filter, args) then changed = true end
  if filter:apply_args(args, true) then changed = true end
  if changed then
    if not temporary then
      filter:save()
    end
  else
    if created then
      ls.printf("Found no filter named '%s'.", filter.name)
      return
    end
  end

  ls.dump(nil, filter)
end

ls.ui = UI.CreateContext("LootSorter")

Command.Event.Attach(Event.Item.Slot, ls.refresh, "LootSorter refresh")
Command.Event.Attach(Event.Item.Update, ls.refresh, "LootSorter refresh")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, ls.variables_loaded, "variable loaded hook")

Library.LibGetOpt.makeslash(filt.Filter:argstring() .. lbag.argstring() .. "f:v", "LootSorter", "ls", ls.slashcommand)
